bl_info = {
    "name": "Natural Stone Wall",
    "author": "Johann Anton Myburgh",
    "version": (1, 1),
    "blender": (2, 80, 0),
    "location": "View3D > Add > Mesh > New Natural Stone Wall",
    "description": "Adds a new Natural Stone Wall",
    "warning": "",
    "doc_url": "",
    "category": "Add Mesh",
}


import bpy
from bpy.types import Operator
from bpy.props import FloatVectorProperty
from bpy.props import IntProperty
from bpy.props import BoolProperty
from bpy.props import StringProperty
from bpy.props import FloatProperty
from bpy_extras.object_utils import AddObjectHelper, object_data_add
from mathutils import Vector
import random
import bmesh
import math
from bpy_extras import object_utils
from mathutils import noise


gridSizeX = 7
gridSizeY = 7
gridSizeZ = 19
cellSize = 0.5
maxExtrudeAmount = 0.4
numSubdivisions = 1
onlyCornersAndIntersections = False
onlyRandomizeInnerVerts = False
firstNoiseOctaveFreq = 0.6
firstNoiseOctaveAmplitude = 0
secondNoiseOctaveFreq = 2
secondNoiseOctaveAmplitude = 0
minBrickWidth = 2 # in number of cells
maxBrickWidth = 3  
minBrickHeight = 4 # in number of cells
maxBrickHeight = 5  
minBrickDepth = 1 # in number of cells
maxBrickDepth = 2  
bevelWidth = 0.05
bevelCount = 5

zCellSize = 1


boxesAspectRatios = []
boxExtentsRight = []
boxExtentsFront = []
boxExtentsBack = []
boxExtentsLeft = []
boxExtentsTop = []
boxExtentsBottom = []

boxPositions = []

numCellsInRowList = []


cellsOfInverseWrapEmptySpaces = []
    
 
grid = []

openCells = []

cylinderBendDegrees = 180
cylinderBendAxis = 0 #  0=X, 1=Y, 2=Z
cylinderBendRadius = 0.2
mustCylinderBend = True
mustCloseCylinderBend = True
mustMakeDome = True




numVerticesInRings = 0
numVerticesInTopFace = 0
numVertsInOneRing = 0
gridVerts = []
gridVertClassifications = [] # 0 = internalVert, 1 = internalFace, 2 = edge, 3 = corner
gridVertEdgeOrientations = [] # 0 = x axis, 1 = y axis, 2 = z axis

def getSubGridIndex(x,y,z):
    return z*((gridSizeX*numSubdivisions+1)*(gridSizeY*numSubdivisions+1))+y*(gridSizeX*numSubdivisions+1)+x

def listContainsArray(list, array):
    for a in list:
        if a[0] == array[0] and a[1] == array[1] and a[2] == array[2]:
            return True
    return False
    
def getAngleOfCoordAlongAxis(lengthOfWallAlongAxis, coordInAxis):
    return math.radians((coordInAxis / lengthOfWallAlongAxis) * cylinderBendDegrees)
    

def randomizeGridVerts():

    
    for z in range(gridSizeZ*numSubdivisions+1):
        for y in range(gridSizeY*numSubdivisions+1):
            for x in range(gridSizeX*numSubdivisions+1):
               
                xCoord = x*cellSize
                yCoord = y*cellSize
                if mustMakeDome:
                    zCoord = z*zCellSize
                else:
                    zCoord = z*cellSize
                
                posVec = Vector((xCoord, yCoord, zCoord))
              
                noiseVec = noise.noise_vector(posVec*firstNoiseOctaveFreq) * firstNoiseOctaveAmplitude + noise.noise_vector(posVec*secondNoiseOctaveFreq) * secondNoiseOctaveAmplitude
                

                if mustCloseCylinderBend:
                    blendFactor =  x - ((gridSizeX - maxBrickWidth - 1) * numSubdivisions)
                    if blendFactor >= 0 and blendFactor < numSubdivisions:
                        
                        xCoordOfWrapCell = -(cellSize*numSubdivisions) + blendFactor * cellSize
                        blendFactor = blendFactor / numSubdivisions
                        
                        posVecOfWrapCell = Vector((xCoordOfWrapCell, yCoord, zCoord))
                        blendNoiseVec = noise.noise_vector(posVecOfWrapCell*firstNoiseOctaveFreq) * firstNoiseOctaveAmplitude + noise.noise_vector(posVecOfWrapCell*secondNoiseOctaveFreq) * secondNoiseOctaveAmplitude
                        noiseVec = (1 - blendFactor) * noiseVec + blendFactor * blendNoiseVec
                    if x >= (gridSizeX - maxBrickWidth) * numSubdivisions:
                        xIndexOfWrapCell = x-((gridSizeX-maxBrickWidth)*numSubdivisions)
                        xCoordOfWrapCell = xIndexOfWrapCell * cellSize
                        
                        
                        posVecOfWrapCell = Vector((xCoordOfWrapCell, yCoord, zCoord))
                        noiseVec = noise.noise_vector(posVecOfWrapCell*firstNoiseOctaveFreq) * firstNoiseOctaveAmplitude + noise.noise_vector(posVecOfWrapCell*secondNoiseOctaveFreq) * secondNoiseOctaveAmplitude
                     
                randXCoord = xCoord+noiseVec.x
                randYCoord = yCoord+noiseVec.y
                randZCoord = zCoord+noiseVec.z

                if not (onlyRandomizeInnerVerts and (x == 0 or x == gridSizeX*numSubdivisions)):
                    xCoord = randXCoord
                
                   
                if mustCloseCylinderBend: 
                    xCoord = randXCoord
                    
                if not (onlyRandomizeInnerVerts and (y == 0 or y == gridSizeY*numSubdivisions)):
                    yCoord = randYCoord
                    
                if not (onlyRandomizeInnerVerts and (z == 0 or z == gridSizeZ*numSubdivisions)):
                    zCoord = randZCoord
                    
               
                gridVerts.append((xCoord,yCoord,zCoord))
                

def getSubGridCoord(m):
    return m * numSubdivisions

def initGridVertsBricksLists():
    for z in range(gridSizeZ*numSubdivisions+1): 
        for y in range(gridSizeY*numSubdivisions+1):
            for x in range(gridSizeX*numSubdivisions+1):
                gridVertClassifications.append(0)  
                gridVertEdgeOrientations.append(-1)
            
def changeGridVertClassification(x,y,z,classification,orientation):
    
    
    if gridVertEdgeOrientations[getSubGridIndex(x,y,z)] > -1:
        if orientation != gridVertEdgeOrientations[getSubGridIndex(x,y,z)]:
            gridVertClassifications[getSubGridIndex(x,y,z)] = 3
    else:
        gridVertEdgeOrientations[getSubGridIndex(x,y,z)] = orientation  
        
    if classification > gridVertClassifications[getSubGridIndex(x,y,z)]:
        gridVertClassifications[getSubGridIndex(x,y,z)] = classification        
    

def populateBricksListXYFace(i,z):
    y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)+1
    for x in range(xStart, xEnd):
        changeGridVertClassification(x,y,z,2,0) 
        
             
    x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    yStart = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    yEnd = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)+1
    for y in range(yStart,yEnd):    
        changeGridVertClassification(x,y,z,2,1) 
        
            
    y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i] +1)
    xStart = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    xEnd = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])-1
    for x in range(xStart, xEnd,-1):
        changeGridVertClassification(x,y,z,2,0)   
        
            
    x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    yStart = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    yEnd = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])-1
    for y in range(yStart , yEnd, -1):    
       changeGridVertClassification(x,y,z,2,1) 
            
def populateBricksListZEdge(i,x,y):
    zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)+1
    for z in range(zStart , zEnd):    
        changeGridVertClassification(x,y,z,2,2) 


def populateBricksLists():
    for i in range(len(boxPositions)):
       
        populateBricksListXYFace(i,getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i]))
        populateBricksListXYFace(i,getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1))
        
        x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
        y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
        populateBricksListZEdge(i,x,y)
        y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
        populateBricksListZEdge(i,x,y)
        
        x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
        y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
        populateBricksListZEdge(i,x,y)
        y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
        populateBricksListZEdge(i,x,y)
        
        
        z = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
        
        y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
        
        x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
        changeGridVertClassification(x,y,z,3,-1) 
        x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
        changeGridVertClassification(x,y,z,3,-1)    
        
        y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
        
        x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
        changeGridVertClassification(x,y,z,3,-1) 
        x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
        changeGridVertClassification(x,y,z,3,-1)     
        
        z = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
        
        y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
        
        x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
        changeGridVertClassification(x,y,z,3,-1) 
        x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
        changeGridVertClassification(x,y,z,3,-1)    
        
        y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
        
        x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
        changeGridVertClassification(x,y,z,3,-1) 
        x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
        changeGridVertClassification(x,y,z,3,-1)      
             
def getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount):

    
    xOffset = 0
    yOffset = 0
    zOffset = 0
    if x == 0: 
        xOffset = -leftExtrudeAmount
    if x == gridSizeX*numSubdivisions:
        xOffset = rightExtrudeAmount
    if z == 0: 
        zOffset = -frontExtrudeAmount
    if z == gridSizeZ*numSubdivisions:
        zOffset = backExtrudeAmount    
    if y == 0: 
        yOffset = -bottomExtrudeAmount
    if y == gridSizeY*numSubdivisions:
        yOffset = topExtrudeAmount    


    return vert[0]+xOffset,vert[1]+yOffset,vert[2]+zOffset


   
def addVertices(i):
    global numVerticesInRings
    global numVerticesInTopFace
    global numVertsInOneRing
    newVerts = []
    # first add horizontal rings from bottom to top.  Each ring starting on left side of front face and ending on right side of left face
    yStart = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    yEnd = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)+1
    
    rightExtrudeAmount = random.random() * maxExtrudeAmount
    leftExtrudeAmount = random.random() * maxExtrudeAmount
    if mustCloseCylinderBend:
        rightExtrudeAmount = 0
        leftExtrudeAmount = 0
    frontExtrudeAmount = random.random() * maxExtrudeAmount
    backExtrudeAmount = random.random() * maxExtrudeAmount
    topExtrudeAmount = random.random() * maxExtrudeAmount
    bottomExtrudeAmount = random.random() * maxExtrudeAmount
    
    for y in range(yStart, yEnd):
        
        z = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
        xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i]) 
        xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)+1 
        for x in range(xStart, xEnd):
            vert = gridVerts[getSubGridIndex(x,y,z)]
            extrudedVert = getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount)
            
            newVerts.append(extrudedVert)
        
        x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
        zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])+1
        zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)+1 
        for z in range(zStart, zEnd):
            vert = gridVerts[getSubGridIndex(x,y,z)]
            extrudedVert = getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount)
            
            newVerts.append(extrudedVert)
                     
        z = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
        xStart = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)-1
        xEnd = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i]) -1
        for x in range(xStart, xEnd, -1):
            vert = gridVerts[getSubGridIndex(x,y,z)]
            extrudedVert = getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount)
            
            newVerts.append(extrudedVert)
                        
        x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
        zStart = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)-1
        zEnd = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i]) 
        for z in range(zStart, zEnd, -1):
            vert = gridVerts[getSubGridIndex(x,y,z)]
            extrudedVert = getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount)
            
            newVerts.append(extrudedVert)
                                
    numVerticesInRings = int(len(newVerts))
    numVertsInOneRing = int(numVerticesInRings / (yEnd - yStart))
    
    # then add verts from top face from left to right and from front to back
    y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])+1 
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    for x in range(xStart, xEnd):
        zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])+1
        zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
        for z in range(zStart, zEnd):
            vert = gridVerts[getSubGridIndex(x,y,z)]
            extrudedVert = getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount)
            
            newVerts.append(extrudedVert)            
    numVerticesInTopFace = int(len(newVerts)-numVerticesInRings)      
            
    # then add verts from bottom face from left to right and from front to back
    y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])+1 
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    for x in range(xStart, xEnd):
        zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])+1
        zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
        for z in range(zStart, zEnd):
            vert = gridVerts[getSubGridIndex(x,y,z)]
            extrudedVert = getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount)
            
            newVerts.append(extrudedVert)   
    
    return newVerts
                        
def getIndexIntoVertList(x,y,z,i):
    
    global numVerticesInRings
    global numVerticesInTopFace
    global numVertsInOneRing
    
    if y == getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1):
        if z > getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i]) and z < getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1):  
            if x > getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i]) and x < getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1):              
                xOffset = x-getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])-1
                zOffset = z-getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])-1
                
                return int(numVerticesInRings + xOffset * (getSubGridCoord(boxExtentsBack[i]+boxExtentsFront[i]+1)-1) + zOffset)
            
    
    if y == getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i]):
        if z > getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i]) and z < getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1):  
            if x > getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i]) and x < getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1):              
                xOffset = x-getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])-1
                zOffset = z-getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])-1
                
                return int(numVerticesInRings + numVerticesInTopFace + xOffset * (getSubGridCoord(boxExtentsBack[i]+boxExtentsFront[i]+1)-1) + zOffset)
    
    
    yOffset = y - getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i]) 
    xOffset = x - getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    zOffset = z - getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    zOffsetFromBack = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1) - z -1
    xOffsetFromRight = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1) - x 
    
    if z == getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i]):
        
        return int(numVertsInOneRing * yOffset + xOffset)
    
    numVertsInFrontRow = getSubGridCoord(boxExtentsLeft[i]+boxExtentsRight[i]+1)+1
    numVertsInRightRow = getSubGridCoord(boxExtentsBack[i]+boxExtentsFront[i]+1)
    numVertsInBackRow = numVertsInFrontRow-1
    
    if x == getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1):
        
        return int(numVertsInOneRing * yOffset + numVertsInFrontRow-1 + zOffset) 
    if z == getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1):
        
        return int(numVertsInOneRing * yOffset + numVertsInFrontRow + numVertsInRightRow + xOffsetFromRight-1)
    
    return int(numVertsInOneRing * yOffset + numVertsInFrontRow + numVertsInRightRow + numVertsInBackRow + zOffsetFromBack)

def createFrontFace(i, startIndex):                    
    faces = []
    z = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    yStart = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    yEnd = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    for y in range(yStart, yEnd):
        for x in range(xStart, xEnd):
            
            faces.append((getIndexIntoVertList(x,y,z,i)+startIndex, getIndexIntoVertList(x,y+1,z,i)+startIndex, getIndexIntoVertList(x+1,y+1,z,i)+startIndex, getIndexIntoVertList(x+1,y,z,i)+startIndex))
            
    return faces    

def createBackFace(i, startIndex):                    
    faces = []
    z = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
    yStart = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    yEnd = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    for y in range(yStart, yEnd):
        for x in range(xStart, xEnd):
            
            faces.append((getIndexIntoVertList(x,y,z,i)+startIndex, getIndexIntoVertList(x+1,y,z,i)+startIndex, getIndexIntoVertList(x+1,y+1,z,i)+startIndex, getIndexIntoVertList(x,y+1,z,i)+startIndex))
            
    return faces    

def createRightFace(i, startIndex):                    
    faces = []
    x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    yStart = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    yEnd = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
    for y in range(yStart, yEnd):
        for z in range(zStart, zEnd):
            
            faces.append((getIndexIntoVertList(x,y,z,i)+startIndex, getIndexIntoVertList(x,y+1,z,i)+startIndex, getIndexIntoVertList(x,y+1,z+1,i)+startIndex, getIndexIntoVertList(x,y,z+1,i)+startIndex))
            
    return faces 

def createLeftFace(i, startIndex):                    
    faces = []
    x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    yStart = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    yEnd = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
    for y in range(yStart, yEnd):
        for z in range(zStart, zEnd):
            
            faces.append((getIndexIntoVertList(x,y,z,i)+startIndex, getIndexIntoVertList(x,y,z+1,i)+startIndex, getIndexIntoVertList(x,y+1,z+1,i)+startIndex, getIndexIntoVertList(x,y+1,z,i)+startIndex))
            
    return faces       

def createBottomFace(i, startIndex):                    
    faces = []
    y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
    for z in range(zStart, zEnd):
        for x in range(xStart, xEnd):
            
            faces.append((getIndexIntoVertList(x,y,z,i)+startIndex, getIndexIntoVertList(x+1,y,z,i)+startIndex, getIndexIntoVertList(x+1,y,z+1,i)+startIndex, getIndexIntoVertList(x,y,z+1,i)+startIndex))
            
    return faces       

def createTopFace(i, startIndex):                    
    faces = []
    y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
    for z in range(zStart, zEnd):
        for x in range(xStart, xEnd):
            
            faces.append((getIndexIntoVertList(x,y,z,i)+startIndex, getIndexIntoVertList(x,y,z+1,i)+startIndex, getIndexIntoVertList(x+1,y,z+1,i)+startIndex, getIndexIntoVertList(x+1,y,z,i)+startIndex))
            
    return faces       


def bendVerts(verts):
    bendedVerts = []
    lengthOfWall = gridSizeX * numSubdivisions * cellSize
    circumfOfDome = (gridSizeY * numSubdivisions * cellSize + cylinderBendRadius) 
    radiusOfDome =  circumfOfDome / math.radians(90) 
    for i in range(len(verts)):
        angleAlongAxis = getAngleOfCoordAlongAxis(lengthOfWall,  verts[i][0])                 
        if mustMakeDome:
            angleForZ = math.radians(90-((verts[i][1] + cylinderBendRadius) / ((gridSizeY*numSubdivisions)*cellSize + cylinderBendRadius)*90))
            distanceFromCenter = (radiusOfDome + verts[i][2]) * math.cos(angleForZ)
            xCoord = distanceFromCenter * math.cos(angleAlongAxis)
            yCoord = distanceFromCenter * math.sin(angleAlongAxis)
            zCoord = (radiusOfDome + verts[i][2]) * math.sin(angleForZ)
        else:
            
            radius = cylinderBendRadius + verts[i][1]
            xCoord = radius * math.cos(angleAlongAxis)
            yCoord = radius * math.sin(angleAlongAxis)
            zCoord = verts[i][2]
        bendedVerts.append((xCoord,yCoord,zCoord))  
    return bendedVerts 

def createFaces(self, context):

    edges = []
    
    vertexList = []
    faceList = []


    populateBricksLists()
    randomizeGridVerts()

    
    if onlyCornersAndIntersections == True:
        straightenEdges()
    for i in range(len(boxPositions)):
     
        verts = []
        verts += addVertices(i)
        if mustCylinderBend:
            verts = bendVerts(verts)
        
        faces = []
        faces += createFrontFace(i, len(vertexList))
        faces += createBackFace(i, len(vertexList))
        faces += createRightFace(i, len(vertexList))
        faces += createLeftFace(i, len(vertexList))
        faces += createBottomFace(i, len(vertexList))
        faces += createTopFace(i, len(vertexList))
        
        faceList += faces
        vertexList += verts
        

    return vertexList, faceList


def straightenEdges():
    global gridVerts
    
    for z in range(gridSizeZ*numSubdivisions+1):   
        for y in range(gridSizeY*numSubdivisions+1):
            for x in range(gridSizeX*numSubdivisions+1):
                if gridVertClassifications[getSubGridIndex(x,y,z)] == 2:
                    firstCorner = (0,0,0)
                    secondCorner = (0,0,0)
                    distanceFromFirstCorner = 0
                    distanceFromFirstCornerToSecond = 0
                    firstCornerDetected = False
                    secondCornerDetected = False
                    if gridVertEdgeOrientations[getSubGridIndex(x,y,z)] == 0:
                        
                        
                        xIndex = x   
                        while xIndex >= 0:
                            if gridVertClassifications[getSubGridIndex(xIndex,y,z)] == 3:
                                firstCornerDetected = True
                                
                                firstCorner = gridVerts[getSubGridIndex(xIndex,y,z)]  
                                distanceFromFirstCorner = x - xIndex  
                                break
                            xIndex -= 1
                        xIndex = x    
                        while xIndex < gridSizeX*numSubdivisions+1:
                            if gridVertClassifications[getSubGridIndex(xIndex,y,z)] == 3:
                                secondCornerDetected = True
                                secondCorner = gridVerts[getSubGridIndex(xIndex,y,z)] 
                                distanceFromFirstCornerToSecond = distanceFromFirstCorner + (xIndex-x)   
                                break
                            xIndex += 1
                          
                        
                    if gridVertEdgeOrientations[getSubGridIndex(x,y,z)] == 1:  
                      
                        yIndex = y 
                        while yIndex >= 0:
                            if gridVertClassifications[getSubGridIndex(x,yIndex,z)] == 3:
                                firstCornerDetected = True
                                firstCorner = gridVerts[getSubGridIndex(x,yIndex,z)] 
                                distanceFromFirstCorner = y - yIndex     
                                break
                            yIndex -= 1
                        yIndex = y   
                        while yIndex < gridSizeY*numSubdivisions+1:
                            if gridVertClassifications[getSubGridIndex(x,yIndex,z)] == 3:
                                secondCornerDetected = True
                                secondCorner = gridVerts[getSubGridIndex(x,yIndex,z)]    
                                distanceFromFirstCornerToSecond = distanceFromFirstCorner + (yIndex-y)
                                break
                            yIndex += 1    

                    if gridVertEdgeOrientations[getSubGridIndex(x,y,z)] == 2:  
      
                        zIndex = z 
                        while zIndex >= 0:
                            if gridVertClassifications[getSubGridIndex(x,y,zIndex)] == 3:
                                firstCornerDetected = True
                                firstCorner = gridVerts[getSubGridIndex(x,y,zIndex)]  
                                distanceFromFirstCorner = z - zIndex    
                                break
                            zIndex -= 1
                        zIndex = z 
                        while zIndex < gridSizeZ*numSubdivisions+1:
                            if gridVertClassifications[getSubGridIndex(x,y,zIndex)] == 3:
                                secondCornerDetected = True
                                secondCorner = gridVerts[getSubGridIndex(x,y,zIndex)]    
                                distanceFromFirstCornerToSecond = distanceFromFirstCorner + (zIndex-z)
                                break
                            zIndex += 1   
                            
                    xDiff = secondCorner[0] - firstCorner[0]
                    yDiff = secondCorner[1] - firstCorner[1]
                    zDiff = secondCorner[2] - firstCorner[2]
    
                        
                    proportionOfTotalDist = distanceFromFirstCorner/distanceFromFirstCornerToSecond
                    interpolatedXPos = firstCorner[0] + proportionOfTotalDist * xDiff
                    interpolatedYPos = firstCorner[1] + proportionOfTotalDist * yDiff 
                    interpolatedZPos = firstCorner[2] + proportionOfTotalDist * zDiff
                    
                    gridVerts[getSubGridIndex(x,y,z)] = (interpolatedXPos,interpolatedYPos,interpolatedZPos)
                   

def expandHorizontally(i):
    global expanded
    isRightColumnClear = True
    isLeftColumnClear = True 
    for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
        for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
            if boxPositions[i][0] + boxExtentsRight[i]+1 < gridSizeX:
                if grid[z][y][boxPositions[i][0] + boxExtentsRight[i]+1] != -1:
                    isRightColumnClear = False
            else: 
                isRightColumnClear = False
    for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
        for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
            if boxPositions[i][0] - boxExtentsLeft[i]-1 >= 0:
                if grid[z][y][boxPositions[i][0] - boxExtentsLeft[i]-1] != -1:
                    isLeftColumnClear = False
            else: 
                isLeftColumnClear = False
    if isRightColumnClear:
        for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
            for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
                x = boxPositions[i][0] + boxExtentsRight[i]+1
                grid[z][y][x] = i
                openCells.remove(z*gridSizeX*gridSizeY + y * gridSizeX + x)
        boxExtentsRight[i] = boxExtentsRight[i] + 1
        expanded = True
    if isLeftColumnClear:
        for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
            for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
                x = boxPositions[i][0] - boxExtentsLeft[i]-1
                grid[z][y][x] = i
                openCells.remove(z*gridSizeX*gridSizeY + y * gridSizeX + x)
        boxExtentsLeft[i] = boxExtentsLeft[i] + 1
        expanded = True
    return expanded  

def expandVertically(i):
    global expanded
    isTopRowClear = True
    isBottomRowClear = True
    for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
        for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
            if boxPositions[i][1] + boxExtentsBottom[i]+1 < gridSizeY:
                if grid[z][boxPositions[i][1] + boxExtentsBottom[i]+1][x] != -1:
                    isBottomRowClear = False
            else:
                isBottomRowClear = False
    for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
        for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
            if boxPositions[i][1] - boxExtentsTop[i]-1 >= 0:
                if grid[z][boxPositions[i][1] - boxExtentsTop[i]-1][x] != -1:
                    isTopRowClear = False
            else:
                isTopRowClear = False   
    if isTopRowClear:
        for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
            for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
                y = boxPositions[i][1] - boxExtentsTop[i]-1
                grid[z][y][x] = i
                openCells.remove(z*gridSizeX*gridSizeY + y*gridSizeX+x)
        boxExtentsTop[i] = boxExtentsTop[i] + 1
        expanded = True
    if isBottomRowClear:
        for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
            for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
                y = boxPositions[i][1] + boxExtentsBottom[i]+1
                grid[z][y][x] = i
                openCells.remove(z*gridSizeX*gridSizeY + y*gridSizeX+x)
        boxExtentsBottom[i] = boxExtentsBottom[i] + 1
        expanded = True
     
        
def expandOnZAxis(i):
    global expanded
    isFrontRowClear = True
    isBackRowRowClear = True
    for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
        for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
            if boxPositions[i][2] - boxExtentsFront[i]-1 >= 0:
                if grid[boxPositions[i][2] - boxExtentsFront[i]-1][y][x] != -1:
                    isFrontRowClear = False
            else:
                isFrontRowClear = False
    for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
        for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
            if boxPositions[i][2] + boxExtentsBack[i]+1 < gridSizeZ:
                if grid[boxPositions[i][2] + boxExtentsBack[i]+1][y][x] != -1:
                    isBackRowRowClear = False
            else:
                isBackRowRowClear = False   
    if isFrontRowClear:
        for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
            for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
                z = boxPositions[i][2] - boxExtentsFront[i]-1
                grid[z][y][x] = i
                openCells.remove(z*gridSizeX*gridSizeY + y*gridSizeX+x)
        boxExtentsFront[i] = boxExtentsFront[i] + 1
        expanded = True
    if isBackRowRowClear:
        for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
            for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
                z = boxPositions[i][2] + boxExtentsBack[i]+1
                grid[z][y][x] = i
                openCells.remove(z*gridSizeX*gridSizeY + y*gridSizeX+x)
        boxExtentsBack[i] = boxExtentsBack[i] + 1
        expanded = True  
      



def buildBricks():
    global expanded


    while len(openCells) > 0:
        randomCell = random.randint(0,len(openCells)-1)
        boxPosZ = int(openCells[randomCell] / (gridSizeY*gridSizeX))
        boxPosY = int( (openCells[randomCell] % (gridSizeY*gridSizeX) ) / gridSizeX)
        boxPosX = int( (openCells[randomCell] % (gridSizeY*gridSizeX) ) % gridSizeX)
        del(openCells[randomCell])
        boxPositions.append([boxPosX,boxPosY,boxPosZ])
        boxExtentsRight.append(0)
        boxExtentsLeft.append(0)
        boxExtentsTop.append(0)
        boxExtentsBottom.append(0)
        boxExtentsFront.append(0)
        boxExtentsBack.append(0)
        
        boxWidth = random.randint(minBrickWidth,maxBrickWidth)
        boxHeight = random.randint(minBrickHeight,maxBrickHeight)
        boxDepth = random.randint(minBrickDepth,maxBrickDepth)
        
        boxesAspectRatios.append([boxWidth, boxDepth, boxHeight])
        
       
        grid[boxPosZ][boxPosY][boxPosX] = len(boxPositions)-1
       # expanded = True
    #    while expanded == True:
     #       expanded = False
       # for i in range(len(boxPositions)):
        expanded = True
        i = len(boxPositions)-1
        while expanded == True:
            expanded = False
            currentWidth = boxExtentsRight[i] + boxExtentsLeft[i] + 1
            currentDepth = boxExtentsBottom[i] + boxExtentsTop[i] + 1
            currentHeight = boxExtentsFront[i] + boxExtentsBack[i] + 1
            currentAspectRatioXZ = float(currentWidth) / float(currentHeight)
            currentAspectRatioYZ = float(currentDepth) / float(currentHeight)
            #if random.randint(0,1) == 0:
         #   if True:
            if currentAspectRatioXZ > float(boxesAspectRatios[i][0]) / float(boxesAspectRatios[i][2]):
                if currentHeight < boxesAspectRatios[i][2]:
                    expandOnZAxis(i)
            else:
                if currentWidth < boxesAspectRatios[i][0]:
                    expandHorizontally(i)
                
            if currentAspectRatioYZ > float(boxesAspectRatios[i][1]) / float(boxesAspectRatios[i][2]):
                if currentHeight < boxesAspectRatios[i][2]:
                    expandOnZAxis(i)
            else:
                if currentDepth < boxesAspectRatios[i][1]:
                    expandVertically(i)
            
        

def add_object(self, context):
    global gridSizeX
    global gridSizeY
    global gridSizeZ
    

    
    global boxesAspectRatios
    global boxExtentsRight
    global boxExtentsFront
    global boxExtentsBack
    global boxExtentsLeft
    global boxExtentsTop
    global boxExtentsBottom

    global boxPositions
    
 
    global grid

    global openCells

    global mustCylinderBend
    global mustCloseCylinderBend 
   
    global gridVerts
    global numSubdivisions
    global onlyRandomizeInnerVerts
    global onlyCornersAndIntersections
    global firstNoiseOctaveFreq
    global firstNoiseOctaveAmplitude
    global secondNoiseOctaveFreq
    global secondNoiseOctaveAmplitude
    global minBrickWidth
    global maxBrickWidth
    global minBrickHeight
    global maxBrickHeight
    global minBrickDepth
    global maxBrickDepth 
    global cellSize
    global maxExtrudeAmount 
    global bevelWidth
    global bevelCount
    global cylinderBendDegrees
    global cylinderBendRadius
    
    global zCellSize

   
    global cellsOfInverseWrapEmptySpaces

    global gridVertClassifications
    global gridVertEdgeOrientations  
    global numCellsInRowList
    global mustMakeDome

    numCellsInRowList = [] 

    cellsOfInverseWrapEmptySpaces = []
    gridSizeX = self.gridDimX
    gridSizeY = self.gridDimY
    gridSizeZ = self.gridDimZ
    
 
    bevelWidth = self.bevelW
    bevelCount = self.bevelCnt
     

    
    numSubdivisions = self.numSubDivs + 1
    if self.onlyCorners:
        numSubdivisions = 1
    cellSize = self.cellDim / numSubdivisions
    onlyRandomizeInnerVerts = self.innerVertsOnly
    
    cylinderBendDegrees = self.cylindricalBendDegrees
    cylinderBendRadius = self.cylindricalBendRadius

    onlyCornersAndIntersections = self.onlyCorners
    firstNoiseOctaveFreq = self.firstNoiseFreq
    firstNoiseOctaveAmplitude = self.firstNoiseAmp
    secondNoiseOctaveFreq = self.secondNoiseFreq
    secondNoiseOctaveAmplitude = self.secondNoiseAmp
    minBrickWidth = self.minBrickW
    maxBrickWidth = self.maxBrickW
    minBrickHeight = self.minBrickH
    maxBrickHeight = self.maxBrickH
    minBrickDepth = self.minBrickD
    maxBrickDepth = self.maxBrickD
    maxExtrudeAmount = self.maxExtrude
    
    zCellSize = self.domeWidth / numSubdivisions
    
    mustCylinderBend = self.bend
    mustCloseCylinderBend = self.closeBend
    mustMakeDome = self.dome
    
    if maxBrickWidth < minBrickWidth:
        dummy = minBrickWidth
        minBrickWidth = maxBrickWidth
        maxBrickWidth = dummy
        
    if maxBrickHeight < minBrickHeight:
        dummy = minBrickHeight
        minBrickHeight = maxBrickHeight
        maxBrickHeight = dummy
        
    if maxBrickDepth < minBrickDepth:
        dummy = minBrickDepth
        minBrickDepth = maxBrickDepth
        maxBrickDepth = dummy
    
    gridSizeY = self.gridDimY
    gridSizeZ = self.gridDimZ    

    # build cylindrical bend wall segment if necessary
    if mustCloseCylinderBend:
        cylinderBendDegrees = 360 + (360 / (gridSizeX - maxBrickWidth)) * maxBrickWidth
        
        
        gridSizeX = maxBrickWidth
    
        

        boxesAspectRatios = []
        boxExtentsRight = []
        boxExtentsFront = []
        boxExtentsBack = []
        boxExtentsLeft = []
        boxExtentsTop = []
        boxExtentsBottom = []

        boxPositions = []
        

        grid = [[[-1 for cell in range(gridSizeX)] for row in range(gridSizeY)] for slice in range(gridSizeZ)]
        
        openCells = list(range(gridSizeX * gridSizeY * gridSizeZ))

        buildBricks()
    
    
        # find bricks bordering right edge of bend segment
    
        allBrickIds = range(len(boxPositions))
          
        brickIDs = []
        for y in range(gridSizeY):
            for z in range(gridSizeZ):
                brickID = grid[z][y][gridSizeX-1]
                if brickID not in brickIDs:
                    brickIDs.append(brickID)
        
    gridSizeX = self.gridDimX
    grid = [[[-1 for cell in range(gridSizeX)] for row in range(gridSizeY)] for slice in range(gridSizeZ)]
    
    
    openCells = list(range(gridSizeX * gridSizeY * gridSizeZ))  
          
      
    if mustCloseCylinderBend:    
        
        for i in brickIDs:
            for z in range(boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
                for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
                    for x in range(boxPositions[i][0] - boxExtentsLeft[i] + gridSizeX - maxBrickWidth, boxPositions[i][0] + boxExtentsRight[i] + gridSizeX - maxBrickWidth + 1 ):
                        indexIntoList = z*gridSizeX*gridSizeY + y*gridSizeX+x
                        if indexIntoList in openCells:
                            openCells.remove(z*gridSizeX*gridSizeY + y*gridSizeX+x)
                        grid[z][y][x] = 1
            
                
        inverseBrickSelection = [i for i in allBrickIds if i not in brickIDs]
        
        for i in inverseBrickSelection:
            for z in range(boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
                for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
                    for x in range(boxPositions[i][0] - boxExtentsLeft[i] , boxPositions[i][0] + boxExtentsRight[i]  + 1 ):
                        indexIntoList = z*gridSizeX*gridSizeY + y*gridSizeX+x
                        if indexIntoList in openCells:
                            openCells.remove(indexIntoList)
                        grid[z][y][x] = 1
                        cellsOfInverseWrapEmptySpaces.append([x,y,z])
                        cellsOfInverseWrapEmptySpaces.append([x+1,y,z])
                        cellsOfInverseWrapEmptySpaces.append([x,y+1,z])
                        cellsOfInverseWrapEmptySpaces.append([x+1,y+1,z])
                        
                        cellsOfInverseWrapEmptySpaces.append([x,y,z+1])
                        cellsOfInverseWrapEmptySpaces.append([x+1,y,z+1])
                        cellsOfInverseWrapEmptySpaces.append([x,y+1,z+1])
                        cellsOfInverseWrapEmptySpaces.append([x+1,y+1,z+1])
                        
            
        
    
    boxPositions = []
    boxExtentsRight = []

    boxExtentsFront = []
    boxExtentsBack = []
    boxExtentsLeft = []
    boxExtentsTop = []
    boxExtentsBottom = []
    boxesAspectRatios = []
    
    buildBricks()
    

    gridVerts = []
    gridVertClassifications = [] 
    gridVertEdgeOrientations = [] 
    initGridVertsBricksLists()
   
    return createFaces(self,context)

class OBJECT_OT_add_object(Operator, AddObjectHelper):
    """创建一个新的天然石墙"""
    bl_idname = "mesh.add_natural_stone_wall"
    bl_label = "Add Natural Stone Wall"
    bl_options = {'REGISTER', 'UNDO'}
    
    cylindricalBendDegrees: FloatProperty(
        name="弯曲角度",
        
        
        default=180.0,
        step=10.0,
        description="圆柱弯曲角度",
    )
   
    gridDimX: IntProperty(
        name="X 单元数量",
        min=1,
        default=17,
        step=1,
        description="Number of cells along X axis",
    
    )
    
    gridDimY: IntProperty(
        name="Y 单元数量",
        min=1,
        default=10,
        step=1,
        description="Number of cells along Y axis",        
    )
    
    gridDimZ: IntProperty(
        name="Z 单元数量",
        min=1,
        default=1,
        step=1,
        description="Number of cells along Z axis",
    )
    
    cellDim: FloatProperty(
        name="单元格大小",
        min=0.00001,
        default=0.5,
        step=25,
        description="Size of one cell in world coords",
    )
    
    numSubDivs: IntProperty(
        name="细分数量",
        min=0,
        default=0,
        step=1,
        description="Number of times each cell must be subdivided",
    )
    
    innerVertsOnly: BoolProperty(
        name="仅随机化内部",
        default=False,
        description="Only randomize vertices that arent on the edge of the object",
    )
    
    onlyCorners: BoolProperty(
        name="仅随机化角点",
        default=False,
        description="Only randomize vertices on corners and intersections of bricks",
    )
    
    bend: BoolProperty(
        name="弯曲",
        default=False,
        description="Bend",
    )
    
    cylindricalBendRadius: FloatProperty(
        name="弯曲半径",
        default=2.5,
        step=0.2,
        description="Cylindrical Bend Radius",
    )
    
    closeBend: BoolProperty(
        name="闭合弯曲",
        default=False,
        description="Close Bend",
    )
    
    dome: BoolProperty(
        name="圆顶",
        default=False,
        description="Bend into dome shape",
    )
    
    domeWidth: FloatProperty(
        name="圆顶宽度",
        min=0,
        default=0.5,
        step=0.1,
        description="Width of one leaf of dome",        
    )
    
    maxExtrude: FloatProperty(
        name="最大挤出高度",
        min=0,
        default=0.15,
        step=10,
        description="Maximum height to extrude exterior faces",        
    )
    
   
    firstNoiseFreq: FloatProperty(
        name="第1次 噪声频率",
        min=0.00001,
        default=0.5,
        step=10,
        description="Primary Noise Layer Frequency",
    )
    
    firstNoiseAmp: FloatProperty(
        name="第1次 噪声振幅",
        min=0.00001,
        default=0.2,
        step=10,
        description="Primary Noise Layer Amplitude",
    )
    
    secondNoiseFreq: FloatProperty(
        name="第2次 噪声频率",
        min=0.00001,
        default=2,
        step=10,
        description="Secondary Noise Layer Frequency",
    )
    
    secondNoiseAmp: FloatProperty(
        name="第2次 噪声振幅",
        min=0.00001,
        default=0,
        step=10,
        description="Secondary Noise Layer Amplitude",
    )
    

    minBrickW: IntProperty(
        name="最小砖X尺寸",
        min=1,
        default=1,
        step=1,
        description="Minimum brick size in X dimension expressed in number of cells",        
    )
    
    maxBrickW: IntProperty(
        name="最大砖X尺寸",
        min=1,
        default=3,
        step=1,
        description="Maximum brick size in X dimension expressed in number of cells",        
    )
    
    minBrickD: IntProperty(
        name="最小砖Y尺寸",
        min=1,
        default=1,
        step=1,
        description="Minimum brick size in Y dimension expressed in number of cells",        
    )
    
    maxBrickD: IntProperty(
        name="最大砖Y尺寸",
        min=1,
        default=3,
        step=1,
        description="Maximum brick size in Y dimension expressed in number of cells",        
    )
    
    minBrickH: IntProperty(
        name="最小砖Z尺寸",
        min=1,
        default=1,
        step=1,
        description="Minimum brick size in Z dimension expressed in number of cells",        
    )
    
    maxBrickH: IntProperty(
        name="最大砖块Z尺寸",
        min=1,
        default=3,
        step=1,
        description="Maximum brick size in Z dimension expressed in number of cells",        
    )
    
    bevelW: FloatProperty(
        name="倒角宽度",
        min=0,
        default=0.05,
        step=1,
        description="Width of Bevel Modifier",        
    )

    bevelCnt: IntProperty(
        name="倒角段数",
        min=0,
        default=5,
        step=1,
        description="Number of Bevel Segments",        
    )
    

    
   
    def execute(self, context):
        
        use_enter_edit_mode = bpy.context.preferences.edit.use_enter_edit_mode
        bpy.context.preferences.edit.use_enter_edit_mode = False

        
        verts_array, faces_array = add_object(self, context)
        
        if bpy.context.mode == "OBJECT":
            if context.selected_objects != [] and context.active_object and \
                (context.active_object.data is not None) and ('Wall' in context.active_object.data.keys()):
                
                obj = context.active_object
                oldmesh = obj.data
                oldmeshname = obj.data.name
                mesh = bpy.data.meshes.new("Wall")
                
                mesh.from_pydata(verts_array, [], faces_array)
               
                obj.data = mesh
                for material in oldmesh.materials:
                    obj.data.materials.append(material)
                bpy.data.meshes.remove(oldmesh)
                obj.data.name = oldmeshname
            else:
                mesh = bpy.data.meshes.new("Wall")
                mesh.from_pydata(verts_array, [], faces_array)
                        
                obj = object_utils.object_data_add(context, mesh, operator=self)
                bpy.ops.object.shade_auto_smooth(angle=math.pi/8)

                for poly in obj.data.polygons:
                    poly.use_smooth = True 
                    
                if bevelWidth > 0 and bevelCount > 0:
                    modifier = obj.modifiers.new(name="Bevel", type='BEVEL')
                    modifier.width = bevelWidth
                    modifier.segments = bevelCount
 

            mesh.update()

        
        
        if bpy.context.mode == "EDIT_MESH":
            
            active_object = context.active_object
            name_active_object = active_object.name
            bpy.ops.object.mode_set(mode='OBJECT')
            mesh = bpy.data.meshes.new("TMP")
            mesh.from_pydata(verts_array, [], faces_array)
            obj = object_utils.object_data_add(context, mesh, operator=self)
            obj.select_set(True)
            active_object.select_set(True)
            bpy.context.view_layer.objects.active = active_object
            bpy.ops.object.join()
            context.active_object.name = name_active_object
            bpy.ops.object.mode_set(mode='EDIT')
        
        if use_enter_edit_mode:
            bpy.ops.object.mode_set(mode = 'EDIT')
        
        # restore pre operator state
        bpy.context.preferences.edit.use_enter_edit_mode = use_enter_edit_mode

        return {'FINISHED'}


# Registration

def add_object_button(self, context):
    self.layout.operator(
        OBJECT_OT_add_object.bl_idname,
        text="Add Natural Stone Wall",
        icon='PLUGIN')


# This allows you to right click on a button and link to documentation
def add_object_manual_map():
    url_manual_prefix = "https://docs.blender.org/manual/en/latest/"
    url_manual_mapping = (
        ("bpy.ops.mesh.add_object", "scene_layout/object/types.html"),
    )
    return url_manual_prefix, url_manual_mapping


def register():
    bpy.utils.register_class(OBJECT_OT_add_object)
    bpy.utils.register_manual_map(add_object_manual_map)
    bpy.types.VIEW3D_MT_mesh_add.append(add_object_button)


def unregister():
    bpy.utils.unregister_class(OBJECT_OT_add_object)
    bpy.utils.unregister_manual_map(add_object_manual_map)
    bpy.types.VIEW3D_MT_mesh_add.remove(add_object_button)


if __name__ == "__main__":
    register()
bl_info = {
    "name": "Natural Stone Wall",
    "author": "Johann Anton Myburgh",
    "version": (1, 1),
    "blender": (2, 80, 0),
    "location": "View3D > Add > Mesh > New Natural Stone Wall",
    "description": "Adds a new Natural Stone Wall",
    "warning": "",
    "doc_url": "",
    "category": "Add Mesh",
}


import bpy
from bpy.types import Operator
from bpy.props import FloatVectorProperty
from bpy.props import IntProperty
from bpy.props import BoolProperty
from bpy.props import StringProperty
from bpy.props import FloatProperty
from bpy_extras.object_utils import AddObjectHelper, object_data_add
from mathutils import Vector
import random
import bmesh
import math
from bpy_extras import object_utils
from mathutils import noise


gridSizeX = 7
gridSizeY = 7
gridSizeZ = 19
cellSize = 0.5
maxExtrudeAmount = 0.4
numSubdivisions = 1
onlyCornersAndIntersections = False
onlyRandomizeInnerVerts = False
firstNoiseOctaveFreq = 0.6
firstNoiseOctaveAmplitude = 0
secondNoiseOctaveFreq = 2
secondNoiseOctaveAmplitude = 0
minBrickWidth = 2 # in number of cells
maxBrickWidth = 3  
minBrickHeight = 4 # in number of cells
maxBrickHeight = 5  
minBrickDepth = 1 # in number of cells
maxBrickDepth = 2  
bevelWidth = 0.05
bevelCount = 5

zCellSize = 1


boxesAspectRatios = []
boxExtentsRight = []
boxExtentsFront = []
boxExtentsBack = []
boxExtentsLeft = []
boxExtentsTop = []
boxExtentsBottom = []

boxPositions = []

numCellsInRowList = []


cellsOfInverseWrapEmptySpaces = []
    
 
grid = []

openCells = []

cylinderBendDegrees = 180
cylinderBendAxis = 0 #  0=X, 1=Y, 2=Z
cylinderBendRadius = 0.2
mustCylinderBend = True
mustCloseCylinderBend = True
mustMakeDome = True




numVerticesInRings = 0
numVerticesInTopFace = 0
numVertsInOneRing = 0
gridVerts = []
gridVertClassifications = [] # 0 = internalVert, 1 = internalFace, 2 = edge, 3 = corner
gridVertEdgeOrientations = [] # 0 = x axis, 1 = y axis, 2 = z axis

def getSubGridIndex(x,y,z):
    return z*((gridSizeX*numSubdivisions+1)*(gridSizeY*numSubdivisions+1))+y*(gridSizeX*numSubdivisions+1)+x

def listContainsArray(list, array):
    for a in list:
        if a[0] == array[0] and a[1] == array[1] and a[2] == array[2]:
            return True
    return False
    
def getAngleOfCoordAlongAxis(lengthOfWallAlongAxis, coordInAxis):
    return math.radians((coordInAxis / lengthOfWallAlongAxis) * cylinderBendDegrees)
    

def randomizeGridVerts():

    
    for z in range(gridSizeZ*numSubdivisions+1):
        for y in range(gridSizeY*numSubdivisions+1):
            for x in range(gridSizeX*numSubdivisions+1):
               
                xCoord = x*cellSize
                yCoord = y*cellSize
                if mustMakeDome:
                    zCoord = z*zCellSize
                else:
                    zCoord = z*cellSize
                
                posVec = Vector((xCoord, yCoord, zCoord))
              
                noiseVec = noise.noise_vector(posVec*firstNoiseOctaveFreq) * firstNoiseOctaveAmplitude + noise.noise_vector(posVec*secondNoiseOctaveFreq) * secondNoiseOctaveAmplitude
                

                if mustCloseCylinderBend:
                    blendFactor =  x - ((gridSizeX - maxBrickWidth - 1) * numSubdivisions)
                    if blendFactor >= 0 and blendFactor < numSubdivisions:
                        
                        xCoordOfWrapCell = -(cellSize*numSubdivisions) + blendFactor * cellSize
                        blendFactor = blendFactor / numSubdivisions
                        
                        posVecOfWrapCell = Vector((xCoordOfWrapCell, yCoord, zCoord))
                        blendNoiseVec = noise.noise_vector(posVecOfWrapCell*firstNoiseOctaveFreq) * firstNoiseOctaveAmplitude + noise.noise_vector(posVecOfWrapCell*secondNoiseOctaveFreq) * secondNoiseOctaveAmplitude
                        noiseVec = (1 - blendFactor) * noiseVec + blendFactor * blendNoiseVec
                    if x >= (gridSizeX - maxBrickWidth) * numSubdivisions:
                        xIndexOfWrapCell = x-((gridSizeX-maxBrickWidth)*numSubdivisions)
                        xCoordOfWrapCell = xIndexOfWrapCell * cellSize
                        
                        
                        posVecOfWrapCell = Vector((xCoordOfWrapCell, yCoord, zCoord))
                        noiseVec = noise.noise_vector(posVecOfWrapCell*firstNoiseOctaveFreq) * firstNoiseOctaveAmplitude + noise.noise_vector(posVecOfWrapCell*secondNoiseOctaveFreq) * secondNoiseOctaveAmplitude
                     
                randXCoord = xCoord+noiseVec.x
                randYCoord = yCoord+noiseVec.y
                randZCoord = zCoord+noiseVec.z

                if not (onlyRandomizeInnerVerts and (x == 0 or x == gridSizeX*numSubdivisions)):
                    xCoord = randXCoord
                
                   
                if mustCloseCylinderBend: 
                    xCoord = randXCoord
                    
                if not (onlyRandomizeInnerVerts and (y == 0 or y == gridSizeY*numSubdivisions)):
                    yCoord = randYCoord
                    
                if not (onlyRandomizeInnerVerts and (z == 0 or z == gridSizeZ*numSubdivisions)):
                    zCoord = randZCoord
                    
               
                gridVerts.append((xCoord,yCoord,zCoord))
                

def getSubGridCoord(m):
    return m * numSubdivisions

def initGridVertsBricksLists():
    for z in range(gridSizeZ*numSubdivisions+1): 
        for y in range(gridSizeY*numSubdivisions+1):
            for x in range(gridSizeX*numSubdivisions+1):
                gridVertClassifications.append(0)  
                gridVertEdgeOrientations.append(-1)
            
def changeGridVertClassification(x,y,z,classification,orientation):
    
    
    if gridVertEdgeOrientations[getSubGridIndex(x,y,z)] > -1:
        if orientation != gridVertEdgeOrientations[getSubGridIndex(x,y,z)]:
            gridVertClassifications[getSubGridIndex(x,y,z)] = 3
    else:
        gridVertEdgeOrientations[getSubGridIndex(x,y,z)] = orientation  
        
    if classification > gridVertClassifications[getSubGridIndex(x,y,z)]:
        gridVertClassifications[getSubGridIndex(x,y,z)] = classification        
    

def populateBricksListXYFace(i,z):
    y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)+1
    for x in range(xStart, xEnd):
        changeGridVertClassification(x,y,z,2,0) 
        
             
    x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    yStart = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    yEnd = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)+1
    for y in range(yStart,yEnd):    
        changeGridVertClassification(x,y,z,2,1) 
        
            
    y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i] +1)
    xStart = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    xEnd = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])-1
    for x in range(xStart, xEnd,-1):
        changeGridVertClassification(x,y,z,2,0)   
        
            
    x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    yStart = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    yEnd = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])-1
    for y in range(yStart , yEnd, -1):    
       changeGridVertClassification(x,y,z,2,1) 
            
def populateBricksListZEdge(i,x,y):
    zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)+1
    for z in range(zStart , zEnd):    
        changeGridVertClassification(x,y,z,2,2) 


def populateBricksLists():
    for i in range(len(boxPositions)):
       
        populateBricksListXYFace(i,getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i]))
        populateBricksListXYFace(i,getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1))
        
        x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
        y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
        populateBricksListZEdge(i,x,y)
        y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
        populateBricksListZEdge(i,x,y)
        
        x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
        y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
        populateBricksListZEdge(i,x,y)
        y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
        populateBricksListZEdge(i,x,y)
        
        
        z = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
        
        y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
        
        x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
        changeGridVertClassification(x,y,z,3,-1) 
        x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
        changeGridVertClassification(x,y,z,3,-1)    
        
        y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
        
        x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
        changeGridVertClassification(x,y,z,3,-1) 
        x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
        changeGridVertClassification(x,y,z,3,-1)     
        
        z = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
        
        y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
        
        x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
        changeGridVertClassification(x,y,z,3,-1) 
        x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
        changeGridVertClassification(x,y,z,3,-1)    
        
        y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
        
        x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
        changeGridVertClassification(x,y,z,3,-1) 
        x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
        changeGridVertClassification(x,y,z,3,-1)      
             
def getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount):

    
    xOffset = 0
    yOffset = 0
    zOffset = 0
    if x == 0: 
        xOffset = -leftExtrudeAmount
    if x == gridSizeX*numSubdivisions:
        xOffset = rightExtrudeAmount
    if z == 0: 
        zOffset = -frontExtrudeAmount
    if z == gridSizeZ*numSubdivisions:
        zOffset = backExtrudeAmount    
    if y == 0: 
        yOffset = -bottomExtrudeAmount
    if y == gridSizeY*numSubdivisions:
        yOffset = topExtrudeAmount    


    return vert[0]+xOffset,vert[1]+yOffset,vert[2]+zOffset


   
def addVertices(i):
    global numVerticesInRings
    global numVerticesInTopFace
    global numVertsInOneRing
    newVerts = []
    # first add horizontal rings from bottom to top.  Each ring starting on left side of front face and ending on right side of left face
    yStart = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    yEnd = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)+1
    
    rightExtrudeAmount = random.random() * maxExtrudeAmount
    leftExtrudeAmount = random.random() * maxExtrudeAmount
    if mustCloseCylinderBend:
        rightExtrudeAmount = 0
        leftExtrudeAmount = 0
    frontExtrudeAmount = random.random() * maxExtrudeAmount
    backExtrudeAmount = random.random() * maxExtrudeAmount
    topExtrudeAmount = random.random() * maxExtrudeAmount
    bottomExtrudeAmount = random.random() * maxExtrudeAmount
    
    for y in range(yStart, yEnd):
        
        z = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
        xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i]) 
        xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)+1 
        for x in range(xStart, xEnd):
            vert = gridVerts[getSubGridIndex(x,y,z)]
            extrudedVert = getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount)
            
            newVerts.append(extrudedVert)
        
        x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
        zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])+1
        zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)+1 
        for z in range(zStart, zEnd):
            vert = gridVerts[getSubGridIndex(x,y,z)]
            extrudedVert = getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount)
            
            newVerts.append(extrudedVert)
                     
        z = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
        xStart = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)-1
        xEnd = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i]) -1
        for x in range(xStart, xEnd, -1):
            vert = gridVerts[getSubGridIndex(x,y,z)]
            extrudedVert = getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount)
            
            newVerts.append(extrudedVert)
                        
        x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
        zStart = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)-1
        zEnd = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i]) 
        for z in range(zStart, zEnd, -1):
            vert = gridVerts[getSubGridIndex(x,y,z)]
            extrudedVert = getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount)
            
            newVerts.append(extrudedVert)
                                
    numVerticesInRings = int(len(newVerts))
    numVertsInOneRing = int(numVerticesInRings / (yEnd - yStart))
    
    # then add verts from top face from left to right and from front to back
    y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])+1 
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    for x in range(xStart, xEnd):
        zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])+1
        zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
        for z in range(zStart, zEnd):
            vert = gridVerts[getSubGridIndex(x,y,z)]
            extrudedVert = getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount)
            
            newVerts.append(extrudedVert)            
    numVerticesInTopFace = int(len(newVerts)-numVerticesInRings)      
            
    # then add verts from bottom face from left to right and from front to back
    y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])+1 
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    for x in range(xStart, xEnd):
        zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])+1
        zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
        for z in range(zStart, zEnd):
            vert = gridVerts[getSubGridIndex(x,y,z)]
            extrudedVert = getExtrudedVert(x,y,z,vert,rightExtrudeAmount,leftExtrudeAmount,frontExtrudeAmount,backExtrudeAmount,topExtrudeAmount,bottomExtrudeAmount)
            
            newVerts.append(extrudedVert)   
    
    return newVerts
                        
def getIndexIntoVertList(x,y,z,i):
    
    global numVerticesInRings
    global numVerticesInTopFace
    global numVertsInOneRing
    
    if y == getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1):
        if z > getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i]) and z < getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1):  
            if x > getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i]) and x < getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1):              
                xOffset = x-getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])-1
                zOffset = z-getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])-1
                
                return int(numVerticesInRings + xOffset * (getSubGridCoord(boxExtentsBack[i]+boxExtentsFront[i]+1)-1) + zOffset)
            
    
    if y == getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i]):
        if z > getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i]) and z < getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1):  
            if x > getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i]) and x < getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1):              
                xOffset = x-getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])-1
                zOffset = z-getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])-1
                
                return int(numVerticesInRings + numVerticesInTopFace + xOffset * (getSubGridCoord(boxExtentsBack[i]+boxExtentsFront[i]+1)-1) + zOffset)
    
    
    yOffset = y - getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i]) 
    xOffset = x - getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    zOffset = z - getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    zOffsetFromBack = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1) - z -1
    xOffsetFromRight = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1) - x 
    
    if z == getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i]):
        
        return int(numVertsInOneRing * yOffset + xOffset)
    
    numVertsInFrontRow = getSubGridCoord(boxExtentsLeft[i]+boxExtentsRight[i]+1)+1
    numVertsInRightRow = getSubGridCoord(boxExtentsBack[i]+boxExtentsFront[i]+1)
    numVertsInBackRow = numVertsInFrontRow-1
    
    if x == getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1):
        
        return int(numVertsInOneRing * yOffset + numVertsInFrontRow-1 + zOffset) 
    if z == getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1):
        
        return int(numVertsInOneRing * yOffset + numVertsInFrontRow + numVertsInRightRow + xOffsetFromRight-1)
    
    return int(numVertsInOneRing * yOffset + numVertsInFrontRow + numVertsInRightRow + numVertsInBackRow + zOffsetFromBack)

def createFrontFace(i, startIndex):                    
    faces = []
    z = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    yStart = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    yEnd = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    for y in range(yStart, yEnd):
        for x in range(xStart, xEnd):
            
            faces.append((getIndexIntoVertList(x,y,z,i)+startIndex, getIndexIntoVertList(x,y+1,z,i)+startIndex, getIndexIntoVertList(x+1,y+1,z,i)+startIndex, getIndexIntoVertList(x+1,y,z,i)+startIndex))
            
    return faces    

def createBackFace(i, startIndex):                    
    faces = []
    z = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
    yStart = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    yEnd = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    for y in range(yStart, yEnd):
        for x in range(xStart, xEnd):
            
            faces.append((getIndexIntoVertList(x,y,z,i)+startIndex, getIndexIntoVertList(x+1,y,z,i)+startIndex, getIndexIntoVertList(x+1,y+1,z,i)+startIndex, getIndexIntoVertList(x,y+1,z,i)+startIndex))
            
    return faces    

def createRightFace(i, startIndex):                    
    faces = []
    x = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    yStart = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    yEnd = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
    for y in range(yStart, yEnd):
        for z in range(zStart, zEnd):
            
            faces.append((getIndexIntoVertList(x,y,z,i)+startIndex, getIndexIntoVertList(x,y+1,z,i)+startIndex, getIndexIntoVertList(x,y+1,z+1,i)+startIndex, getIndexIntoVertList(x,y,z+1,i)+startIndex))
            
    return faces 

def createLeftFace(i, startIndex):                    
    faces = []
    x = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    yStart = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    yEnd = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
    for y in range(yStart, yEnd):
        for z in range(zStart, zEnd):
            
            faces.append((getIndexIntoVertList(x,y,z,i)+startIndex, getIndexIntoVertList(x,y,z+1,i)+startIndex, getIndexIntoVertList(x,y+1,z+1,i)+startIndex, getIndexIntoVertList(x,y+1,z,i)+startIndex))
            
    return faces       

def createBottomFace(i, startIndex):                    
    faces = []
    y = getSubGridCoord(boxPositions[i][1] - boxExtentsTop[i])
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
    for z in range(zStart, zEnd):
        for x in range(xStart, xEnd):
            
            faces.append((getIndexIntoVertList(x,y,z,i)+startIndex, getIndexIntoVertList(x+1,y,z,i)+startIndex, getIndexIntoVertList(x+1,y,z+1,i)+startIndex, getIndexIntoVertList(x,y,z+1,i)+startIndex))
            
    return faces       

def createTopFace(i, startIndex):                    
    faces = []
    y = getSubGridCoord(boxPositions[i][1] + boxExtentsBottom[i]+1)
    xStart = getSubGridCoord(boxPositions[i][0] - boxExtentsLeft[i])
    xEnd = getSubGridCoord(boxPositions[i][0] + boxExtentsRight[i]+1)
    zStart = getSubGridCoord(boxPositions[i][2] - boxExtentsFront[i])
    zEnd = getSubGridCoord(boxPositions[i][2] + boxExtentsBack[i]+1)
    for z in range(zStart, zEnd):
        for x in range(xStart, xEnd):
            
            faces.append((getIndexIntoVertList(x,y,z,i)+startIndex, getIndexIntoVertList(x,y,z+1,i)+startIndex, getIndexIntoVertList(x+1,y,z+1,i)+startIndex, getIndexIntoVertList(x+1,y,z,i)+startIndex))
            
    return faces       


def bendVerts(verts):
    bendedVerts = []
    lengthOfWall = gridSizeX * numSubdivisions * cellSize
    circumfOfDome = (gridSizeY * numSubdivisions * cellSize + cylinderBendRadius) 
    radiusOfDome =  circumfOfDome / math.radians(90) 
    for i in range(len(verts)):
        angleAlongAxis = getAngleOfCoordAlongAxis(lengthOfWall,  verts[i][0])                 
        if mustMakeDome:
            angleForZ = math.radians(90-((verts[i][1] + cylinderBendRadius) / ((gridSizeY*numSubdivisions)*cellSize + cylinderBendRadius)*90))
            distanceFromCenter = (radiusOfDome + verts[i][2]) * math.cos(angleForZ)
            xCoord = distanceFromCenter * math.cos(angleAlongAxis)
            yCoord = distanceFromCenter * math.sin(angleAlongAxis)
            zCoord = (radiusOfDome + verts[i][2]) * math.sin(angleForZ)
        else:
            
            radius = cylinderBendRadius + verts[i][1]
            xCoord = radius * math.cos(angleAlongAxis)
            yCoord = radius * math.sin(angleAlongAxis)
            zCoord = verts[i][2]
        bendedVerts.append((xCoord,yCoord,zCoord))  
    return bendedVerts 

def createFaces(self, context):

    edges = []
    
    vertexList = []
    faceList = []


    populateBricksLists()
    randomizeGridVerts()

    
    if onlyCornersAndIntersections == True:
        straightenEdges()
    for i in range(len(boxPositions)):
     
        verts = []
        verts += addVertices(i)
        if mustCylinderBend:
            verts = bendVerts(verts)
        
        faces = []
        faces += createFrontFace(i, len(vertexList))
        faces += createBackFace(i, len(vertexList))
        faces += createRightFace(i, len(vertexList))
        faces += createLeftFace(i, len(vertexList))
        faces += createBottomFace(i, len(vertexList))
        faces += createTopFace(i, len(vertexList))
        
        faceList += faces
        vertexList += verts
        

    return vertexList, faceList


def straightenEdges():
    global gridVerts
    
    for z in range(gridSizeZ*numSubdivisions+1):   
        for y in range(gridSizeY*numSubdivisions+1):
            for x in range(gridSizeX*numSubdivisions+1):
                if gridVertClassifications[getSubGridIndex(x,y,z)] == 2:
                    firstCorner = (0,0,0)
                    secondCorner = (0,0,0)
                    distanceFromFirstCorner = 0
                    distanceFromFirstCornerToSecond = 0
                    firstCornerDetected = False
                    secondCornerDetected = False
                    if gridVertEdgeOrientations[getSubGridIndex(x,y,z)] == 0:
                        
                        
                        xIndex = x   
                        while xIndex >= 0:
                            if gridVertClassifications[getSubGridIndex(xIndex,y,z)] == 3:
                                firstCornerDetected = True
                                
                                firstCorner = gridVerts[getSubGridIndex(xIndex,y,z)]  
                                distanceFromFirstCorner = x - xIndex  
                                break
                            xIndex -= 1
                        xIndex = x    
                        while xIndex < gridSizeX*numSubdivisions+1:
                            if gridVertClassifications[getSubGridIndex(xIndex,y,z)] == 3:
                                secondCornerDetected = True
                                secondCorner = gridVerts[getSubGridIndex(xIndex,y,z)] 
                                distanceFromFirstCornerToSecond = distanceFromFirstCorner + (xIndex-x)   
                                break
                            xIndex += 1
                          
                        
                    if gridVertEdgeOrientations[getSubGridIndex(x,y,z)] == 1:  
                      
                        yIndex = y 
                        while yIndex >= 0:
                            if gridVertClassifications[getSubGridIndex(x,yIndex,z)] == 3:
                                firstCornerDetected = True
                                firstCorner = gridVerts[getSubGridIndex(x,yIndex,z)] 
                                distanceFromFirstCorner = y - yIndex     
                                break
                            yIndex -= 1
                        yIndex = y   
                        while yIndex < gridSizeY*numSubdivisions+1:
                            if gridVertClassifications[getSubGridIndex(x,yIndex,z)] == 3:
                                secondCornerDetected = True
                                secondCorner = gridVerts[getSubGridIndex(x,yIndex,z)]    
                                distanceFromFirstCornerToSecond = distanceFromFirstCorner + (yIndex-y)
                                break
                            yIndex += 1    

                    if gridVertEdgeOrientations[getSubGridIndex(x,y,z)] == 2:  
      
                        zIndex = z 
                        while zIndex >= 0:
                            if gridVertClassifications[getSubGridIndex(x,y,zIndex)] == 3:
                                firstCornerDetected = True
                                firstCorner = gridVerts[getSubGridIndex(x,y,zIndex)]  
                                distanceFromFirstCorner = z - zIndex    
                                break
                            zIndex -= 1
                        zIndex = z 
                        while zIndex < gridSizeZ*numSubdivisions+1:
                            if gridVertClassifications[getSubGridIndex(x,y,zIndex)] == 3:
                                secondCornerDetected = True
                                secondCorner = gridVerts[getSubGridIndex(x,y,zIndex)]    
                                distanceFromFirstCornerToSecond = distanceFromFirstCorner + (zIndex-z)
                                break
                            zIndex += 1   
                            
                    xDiff = secondCorner[0] - firstCorner[0]
                    yDiff = secondCorner[1] - firstCorner[1]
                    zDiff = secondCorner[2] - firstCorner[2]
    
                        
                    proportionOfTotalDist = distanceFromFirstCorner/distanceFromFirstCornerToSecond
                    interpolatedXPos = firstCorner[0] + proportionOfTotalDist * xDiff
                    interpolatedYPos = firstCorner[1] + proportionOfTotalDist * yDiff 
                    interpolatedZPos = firstCorner[2] + proportionOfTotalDist * zDiff
                    
                    gridVerts[getSubGridIndex(x,y,z)] = (interpolatedXPos,interpolatedYPos,interpolatedZPos)
                   

def expandHorizontally(i):
    global expanded
    isRightColumnClear = True
    isLeftColumnClear = True 
    for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
        for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
            if boxPositions[i][0] + boxExtentsRight[i]+1 < gridSizeX:
                if grid[z][y][boxPositions[i][0] + boxExtentsRight[i]+1] != -1:
                    isRightColumnClear = False
            else: 
                isRightColumnClear = False
    for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
        for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
            if boxPositions[i][0] - boxExtentsLeft[i]-1 >= 0:
                if grid[z][y][boxPositions[i][0] - boxExtentsLeft[i]-1] != -1:
                    isLeftColumnClear = False
            else: 
                isLeftColumnClear = False
    if isRightColumnClear:
        for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
            for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
                x = boxPositions[i][0] + boxExtentsRight[i]+1
                grid[z][y][x] = i
                openCells.remove(z*gridSizeX*gridSizeY + y * gridSizeX + x)
        boxExtentsRight[i] = boxExtentsRight[i] + 1
        expanded = True
    if isLeftColumnClear:
        for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
            for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
                x = boxPositions[i][0] - boxExtentsLeft[i]-1
                grid[z][y][x] = i
                openCells.remove(z*gridSizeX*gridSizeY + y * gridSizeX + x)
        boxExtentsLeft[i] = boxExtentsLeft[i] + 1
        expanded = True
    return expanded  

def expandVertically(i):
    global expanded
    isTopRowClear = True
    isBottomRowClear = True
    for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
        for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
            if boxPositions[i][1] + boxExtentsBottom[i]+1 < gridSizeY:
                if grid[z][boxPositions[i][1] + boxExtentsBottom[i]+1][x] != -1:
                    isBottomRowClear = False
            else:
                isBottomRowClear = False
    for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
        for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
            if boxPositions[i][1] - boxExtentsTop[i]-1 >= 0:
                if grid[z][boxPositions[i][1] - boxExtentsTop[i]-1][x] != -1:
                    isTopRowClear = False
            else:
                isTopRowClear = False   
    if isTopRowClear:
        for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
            for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
                y = boxPositions[i][1] - boxExtentsTop[i]-1
                grid[z][y][x] = i
                openCells.remove(z*gridSizeX*gridSizeY + y*gridSizeX+x)
        boxExtentsTop[i] = boxExtentsTop[i] + 1
        expanded = True
    if isBottomRowClear:
        for z in range (boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
            for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
                y = boxPositions[i][1] + boxExtentsBottom[i]+1
                grid[z][y][x] = i
                openCells.remove(z*gridSizeX*gridSizeY + y*gridSizeX+x)
        boxExtentsBottom[i] = boxExtentsBottom[i] + 1
        expanded = True
     
        
def expandOnZAxis(i):
    global expanded
    isFrontRowClear = True
    isBackRowRowClear = True
    for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
        for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
            if boxPositions[i][2] - boxExtentsFront[i]-1 >= 0:
                if grid[boxPositions[i][2] - boxExtentsFront[i]-1][y][x] != -1:
                    isFrontRowClear = False
            else:
                isFrontRowClear = False
    for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
        for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
            if boxPositions[i][2] + boxExtentsBack[i]+1 < gridSizeZ:
                if grid[boxPositions[i][2] + boxExtentsBack[i]+1][y][x] != -1:
                    isBackRowRowClear = False
            else:
                isBackRowRowClear = False   
    if isFrontRowClear:
        for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
            for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
                z = boxPositions[i][2] - boxExtentsFront[i]-1
                grid[z][y][x] = i
                openCells.remove(z*gridSizeX*gridSizeY + y*gridSizeX+x)
        boxExtentsFront[i] = boxExtentsFront[i] + 1
        expanded = True
    if isBackRowRowClear:
        for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
            for x in range(boxPositions[i][0] - boxExtentsLeft[i], boxPositions[i][0] + boxExtentsRight[i]+1):
                z = boxPositions[i][2] + boxExtentsBack[i]+1
                grid[z][y][x] = i
                openCells.remove(z*gridSizeX*gridSizeY + y*gridSizeX+x)
        boxExtentsBack[i] = boxExtentsBack[i] + 1
        expanded = True  
      



def buildBricks():
    global expanded


    while len(openCells) > 0:
        randomCell = random.randint(0,len(openCells)-1)
        boxPosZ = int(openCells[randomCell] / (gridSizeY*gridSizeX))
        boxPosY = int( (openCells[randomCell] % (gridSizeY*gridSizeX) ) / gridSizeX)
        boxPosX = int( (openCells[randomCell] % (gridSizeY*gridSizeX) ) % gridSizeX)
        del(openCells[randomCell])
        boxPositions.append([boxPosX,boxPosY,boxPosZ])
        boxExtentsRight.append(0)
        boxExtentsLeft.append(0)
        boxExtentsTop.append(0)
        boxExtentsBottom.append(0)
        boxExtentsFront.append(0)
        boxExtentsBack.append(0)
        
        boxWidth = random.randint(minBrickWidth,maxBrickWidth)
        boxHeight = random.randint(minBrickHeight,maxBrickHeight)
        boxDepth = random.randint(minBrickDepth,maxBrickDepth)
        
        boxesAspectRatios.append([boxWidth, boxDepth, boxHeight])
        
       
        grid[boxPosZ][boxPosY][boxPosX] = len(boxPositions)-1
       # expanded = True
    #    while expanded == True:
     #       expanded = False
       # for i in range(len(boxPositions)):
        expanded = True
        i = len(boxPositions)-1
        while expanded == True:
            expanded = False
            currentWidth = boxExtentsRight[i] + boxExtentsLeft[i] + 1
            currentDepth = boxExtentsBottom[i] + boxExtentsTop[i] + 1
            currentHeight = boxExtentsFront[i] + boxExtentsBack[i] + 1
            currentAspectRatioXZ = float(currentWidth) / float(currentHeight)
            currentAspectRatioYZ = float(currentDepth) / float(currentHeight)
            #if random.randint(0,1) == 0:
         #   if True:
            if currentAspectRatioXZ > float(boxesAspectRatios[i][0]) / float(boxesAspectRatios[i][2]):
                if currentHeight < boxesAspectRatios[i][2]:
                    expandOnZAxis(i)
            else:
                if currentWidth < boxesAspectRatios[i][0]:
                    expandHorizontally(i)
                
            if currentAspectRatioYZ > float(boxesAspectRatios[i][1]) / float(boxesAspectRatios[i][2]):
                if currentHeight < boxesAspectRatios[i][2]:
                    expandOnZAxis(i)
            else:
                if currentDepth < boxesAspectRatios[i][1]:
                    expandVertically(i)
            
        

def add_object(self, context):
    global gridSizeX
    global gridSizeY
    global gridSizeZ
    

    
    global boxesAspectRatios
    global boxExtentsRight
    global boxExtentsFront
    global boxExtentsBack
    global boxExtentsLeft
    global boxExtentsTop
    global boxExtentsBottom

    global boxPositions
    
 
    global grid

    global openCells

    global mustCylinderBend
    global mustCloseCylinderBend 
   
    global gridVerts
    global numSubdivisions
    global onlyRandomizeInnerVerts
    global onlyCornersAndIntersections
    global firstNoiseOctaveFreq
    global firstNoiseOctaveAmplitude
    global secondNoiseOctaveFreq
    global secondNoiseOctaveAmplitude
    global minBrickWidth
    global maxBrickWidth
    global minBrickHeight
    global maxBrickHeight
    global minBrickDepth
    global maxBrickDepth 
    global cellSize
    global maxExtrudeAmount 
    global bevelWidth
    global bevelCount
    global cylinderBendDegrees
    global cylinderBendRadius
    
    global zCellSize

   
    global cellsOfInverseWrapEmptySpaces

    global gridVertClassifications
    global gridVertEdgeOrientations  
    global numCellsInRowList
    global mustMakeDome

    numCellsInRowList = [] 

    cellsOfInverseWrapEmptySpaces = []
    gridSizeX = self.gridDimX
    gridSizeY = self.gridDimY
    gridSizeZ = self.gridDimZ
    
 
    bevelWidth = self.bevelW
    bevelCount = self.bevelCnt
     

    
    numSubdivisions = self.numSubDivs + 1
    if self.onlyCorners:
        numSubdivisions = 1
    cellSize = self.cellDim / numSubdivisions
    onlyRandomizeInnerVerts = self.innerVertsOnly
    
    cylinderBendDegrees = self.cylindricalBendDegrees
    cylinderBendRadius = self.cylindricalBendRadius

    onlyCornersAndIntersections = self.onlyCorners
    firstNoiseOctaveFreq = self.firstNoiseFreq
    firstNoiseOctaveAmplitude = self.firstNoiseAmp
    secondNoiseOctaveFreq = self.secondNoiseFreq
    secondNoiseOctaveAmplitude = self.secondNoiseAmp
    minBrickWidth = self.minBrickW
    maxBrickWidth = self.maxBrickW
    minBrickHeight = self.minBrickH
    maxBrickHeight = self.maxBrickH
    minBrickDepth = self.minBrickD
    maxBrickDepth = self.maxBrickD
    maxExtrudeAmount = self.maxExtrude
    
    zCellSize = self.domeWidth / numSubdivisions
    
    mustCylinderBend = self.bend
    mustCloseCylinderBend = self.closeBend
    mustMakeDome = self.dome
    
    if maxBrickWidth < minBrickWidth:
        dummy = minBrickWidth
        minBrickWidth = maxBrickWidth
        maxBrickWidth = dummy
        
    if maxBrickHeight < minBrickHeight:
        dummy = minBrickHeight
        minBrickHeight = maxBrickHeight
        maxBrickHeight = dummy
        
    if maxBrickDepth < minBrickDepth:
        dummy = minBrickDepth
        minBrickDepth = maxBrickDepth
        maxBrickDepth = dummy
    
    gridSizeY = self.gridDimY
    gridSizeZ = self.gridDimZ    

    # build cylindrical bend wall segment if necessary
    if mustCloseCylinderBend:
        cylinderBendDegrees = 360 + (360 / (gridSizeX - maxBrickWidth)) * maxBrickWidth
        
        
        gridSizeX = maxBrickWidth
    
        

        boxesAspectRatios = []
        boxExtentsRight = []
        boxExtentsFront = []
        boxExtentsBack = []
        boxExtentsLeft = []
        boxExtentsTop = []
        boxExtentsBottom = []

        boxPositions = []
        

        grid = [[[-1 for cell in range(gridSizeX)] for row in range(gridSizeY)] for slice in range(gridSizeZ)]
        
        openCells = list(range(gridSizeX * gridSizeY * gridSizeZ))

        buildBricks()
    
    
        # find bricks bordering right edge of bend segment
    
        allBrickIds = range(len(boxPositions))
          
        brickIDs = []
        for y in range(gridSizeY):
            for z in range(gridSizeZ):
                brickID = grid[z][y][gridSizeX-1]
                if brickID not in brickIDs:
                    brickIDs.append(brickID)
        
    gridSizeX = self.gridDimX
    grid = [[[-1 for cell in range(gridSizeX)] for row in range(gridSizeY)] for slice in range(gridSizeZ)]
    
    
    openCells = list(range(gridSizeX * gridSizeY * gridSizeZ))  
          
      
    if mustCloseCylinderBend:    
        
        for i in brickIDs:
            for z in range(boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
                for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
                    for x in range(boxPositions[i][0] - boxExtentsLeft[i] + gridSizeX - maxBrickWidth, boxPositions[i][0] + boxExtentsRight[i] + gridSizeX - maxBrickWidth + 1 ):
                        indexIntoList = z*gridSizeX*gridSizeY + y*gridSizeX+x
                        if indexIntoList in openCells:
                            openCells.remove(z*gridSizeX*gridSizeY + y*gridSizeX+x)
                        grid[z][y][x] = 1
            
                
        inverseBrickSelection = [i for i in allBrickIds if i not in brickIDs]
        
        for i in inverseBrickSelection:
            for z in range(boxPositions[i][2] - boxExtentsFront[i], boxPositions[i][2] + boxExtentsBack[i]+1):
                for y in range(boxPositions[i][1] - boxExtentsTop[i], boxPositions[i][1] + boxExtentsBottom[i]+1):
                    for x in range(boxPositions[i][0] - boxExtentsLeft[i] , boxPositions[i][0] + boxExtentsRight[i]  + 1 ):
                        indexIntoList = z*gridSizeX*gridSizeY + y*gridSizeX+x
                        if indexIntoList in openCells:
                            openCells.remove(indexIntoList)
                        grid[z][y][x] = 1
                        cellsOfInverseWrapEmptySpaces.append([x,y,z])
                        cellsOfInverseWrapEmptySpaces.append([x+1,y,z])
                        cellsOfInverseWrapEmptySpaces.append([x,y+1,z])
                        cellsOfInverseWrapEmptySpaces.append([x+1,y+1,z])
                        
                        cellsOfInverseWrapEmptySpaces.append([x,y,z+1])
                        cellsOfInverseWrapEmptySpaces.append([x+1,y,z+1])
                        cellsOfInverseWrapEmptySpaces.append([x,y+1,z+1])
                        cellsOfInverseWrapEmptySpaces.append([x+1,y+1,z+1])
                        
            
        
    
    boxPositions = []
    boxExtentsRight = []

    boxExtentsFront = []
    boxExtentsBack = []
    boxExtentsLeft = []
    boxExtentsTop = []
    boxExtentsBottom = []
    boxesAspectRatios = []
    
    buildBricks()
    

    gridVerts = []
    gridVertClassifications = [] 
    gridVertEdgeOrientations = [] 
    initGridVertsBricksLists()
   
    return createFaces(self,context)

class OBJECT_OT_add_object(Operator, AddObjectHelper):
    """Create a new Natural Stone Wall"""
    bl_idname = "mesh.add_natural_stone_wall"
    bl_label = "Add Natural Stone Wall"
    bl_options = {'REGISTER', 'UNDO'}
    
    cylindricalBendDegrees: FloatProperty(
        name="弯曲角度",
        
        
        default=180.0,
        step=10.0,
        description="Cylindrical Bend Degrees",
    )
   
    gridDimX: IntProperty(
        name="X 单元格数量",
        min=1,
        default=17,
        step=1,
        description="Number of cells along X axis",
    
    )
    
    gridDimY: IntProperty(
        name="Y 单元格数量",
        min=1,
        default=10,
        step=1,
        description="Number of cells along Y axis",        
    )
    
    gridDimZ: IntProperty(
        name="Z 单元格数量",
        min=1,
        default=1,
        step=1,
        description="Number of cells along Z axis",
    )
    
    cellDim: FloatProperty(
        name="单元格大小",
        min=0.00001,
        default=0.5,
        step=25,
        description="Size of one cell in world coords",
    )
    
    numSubDivs: IntProperty(
        name="细分数量",
        min=0,
        default=0,
        step=1,
        description="Number of times each cell must be subdivided",
    )
    
    innerVertsOnly: BoolProperty(
        name="仅随机化内部",
        default=False,
        description="Only randomize vertices that arent on the edge of the object",
    )
    
    onlyCorners: BoolProperty(
        name="仅随机化角点",
        default=False,
        description="Only randomize vertices on corners and intersections of bricks",
    )
    
    bend: BoolProperty(
        name="弯曲",
        default=False,
        description="Bend",
    )
    
    cylindricalBendRadius: FloatProperty(
        name="弯曲半径",
        default=2.5,
        step=0.2,
        description="Cylindrical Bend Radius",
    )
    
    closeBend: BoolProperty(
        name="闭合半径",
        default=False,
        description="Close Bend",
    )
    
    dome: BoolProperty(
        name="圆顶",
        default=False,
        description="Bend into dome shape",
    )
    
    domeWidth: FloatProperty(
        name="圆顶宽度",
        min=0,
        default=0.5,
        step=0.1,
        description="Width of one leaf of dome",        
    )
    
    maxExtrude: FloatProperty(
        name="最大挤出高度",
        min=0,
        default=0.15,
        step=10,
        description="Maximum height to extrude exterior faces",        
    )
    
   
    firstNoiseFreq: FloatProperty(
        name="第1次 噪声频率",
        min=0.00001,
        default=0.5,
        step=10,
        description="Primary Noise Layer Frequency",
    )
    
    firstNoiseAmp: FloatProperty(
        name="第1次 噪声振幅",
        min=0.00001,
        default=0.2,
        step=10,
        description="Primary Noise Layer Amplitude",
    )
    
    secondNoiseFreq: FloatProperty(
        name="第2次 噪声频率",
        min=0.00001,
        default=2,
        step=10,
        description="Secondary Noise Layer Frequency",
    )
    
    secondNoiseAmp: FloatProperty(
        name="第2次 噪声振幅",
        min=0.00001,
        default=0,
        step=10,
        description="Secondary Noise Layer Amplitude",
    )
    

    minBrickW: IntProperty(
        name="最小砖X尺寸",
        min=1,
        default=1,
        step=1,
        description="Minimum brick size in X dimension expressed in number of cells",        
    )
    
    maxBrickW: IntProperty(
        name="最大砖X尺寸",
        min=1,
        default=3,
        step=1,
        description="Maximum brick size in X dimension expressed in number of cells",        
    )
    
    minBrickD: IntProperty(
        name="最小砖Y尺寸",
        min=1,
        default=1,
        step=1,
        description="Minimum brick size in Y dimension expressed in number of cells",        
    )
    
    maxBrickD: IntProperty(
        name="最大砖Y尺寸",
        min=1,
        default=3,
        step=1,
        description="Maximum brick size in Y dimension expressed in number of cells",        
    )
    
    minBrickH: IntProperty(
        name="最小砖Z尺寸",
        min=1,
        default=1,
        step=1,
        description="Minimum brick size in Z dimension expressed in number of cells",        
    )
    
    maxBrickH: IntProperty(
        name="最大砖块Z尺寸",
        min=1,
        default=3,
        step=1,
        description="Maximum brick size in Z dimension expressed in number of cells",        
    )
    
    bevelW: FloatProperty(
        name="倒角宽度",
        min=0,
        default=0.05,
        step=1,
        description="Width of Bevel Modifier",        
    )

    bevelCnt: IntProperty(
        name="倒角段数",
        min=0,
        default=5,
        step=1,
        description="Number of Bevel Segments",        
    )
    

    
   
    def execute(self, context):
        
        use_enter_edit_mode = bpy.context.preferences.edit.use_enter_edit_mode
        bpy.context.preferences.edit.use_enter_edit_mode = False

        
        verts_array, faces_array = add_object(self, context)
        
        if bpy.context.mode == "OBJECT":
            if context.selected_objects != [] and context.active_object and \
                (context.active_object.data is not None) and ('Wall' in context.active_object.data.keys()):
                
                obj = context.active_object
                oldmesh = obj.data
                oldmeshname = obj.data.name
                mesh = bpy.data.meshes.new("Wall")
                
                mesh.from_pydata(verts_array, [], faces_array)
               
                obj.data = mesh
                for material in oldmesh.materials:
                    obj.data.materials.append(material)
                bpy.data.meshes.remove(oldmesh)
                obj.data.name = oldmeshname
            else:
                mesh = bpy.data.meshes.new("Wall")
                mesh.from_pydata(verts_array, [], faces_array)
                        
                obj = object_utils.object_data_add(context, mesh, operator=self)
                bpy.ops.object.shade_auto_smooth(angle=math.pi/8)

                for poly in obj.data.polygons:
                    poly.use_smooth = True 
                    
                if bevelWidth > 0 and bevelCount > 0:
                    modifier = obj.modifiers.new(name="Bevel", type='BEVEL')
                    modifier.width = bevelWidth
                    modifier.segments = bevelCount
 

            mesh.update()

        
        
        if bpy.context.mode == "EDIT_MESH":
            
            active_object = context.active_object
            name_active_object = active_object.name
            bpy.ops.object.mode_set(mode='OBJECT')
            mesh = bpy.data.meshes.new("TMP")
            mesh.from_pydata(verts_array, [], faces_array)
            obj = object_utils.object_data_add(context, mesh, operator=self)
            obj.select_set(True)
            active_object.select_set(True)
            bpy.context.view_layer.objects.active = active_object
            bpy.ops.object.join()
            context.active_object.name = name_active_object
            bpy.ops.object.mode_set(mode='EDIT')
        
        if use_enter_edit_mode:
            bpy.ops.object.mode_set(mode = 'EDIT')
        
        # restore pre operator state
        bpy.context.preferences.edit.use_enter_edit_mode = use_enter_edit_mode

        return {'FINISHED'}


# Registration

def add_object_button(self, context):
    self.layout.operator(
        OBJECT_OT_add_object.bl_idname,
        text="Add Natural Stone Wall",
        icon='PLUGIN')


# This allows you to right click on a button and link to documentation
def add_object_manual_map():
    url_manual_prefix = "https://docs.blender.org/manual/en/latest/"
    url_manual_mapping = (
        ("bpy.ops.mesh.add_object", "scene_layout/object/types.html"),
    )
    return url_manual_prefix, url_manual_mapping


def register():
    bpy.utils.register_class(OBJECT_OT_add_object)
    bpy.utils.register_manual_map(add_object_manual_map)
    bpy.types.VIEW3D_MT_mesh_add.append(add_object_button)


def unregister():
    bpy.utils.unregister_class(OBJECT_OT_add_object)
    bpy.utils.unregister_manual_map(add_object_manual_map)
    bpy.types.VIEW3D_MT_mesh_add.remove(add_object_button)


if __name__ == "__main__":
    register()
